// GML_to_DXF by Golan Levin
// golan at flong dot com
// Version 1.0: 7 January, 2010
//
// Simple GML file conversion program with rudimentary display.
// Converts GML (spec 0.1b) to DXF or CSV.
// More information on GML: http://www.000000book.com/
//
// Loads a GML file and exports a single DXF file, 
// when the user presses the 'd' key. 
// Also, exports CSV files (one per stroke in the tag!)
// when the user presses the 'c' key.
//
// Files appear in the "Sketch" folder.
// Press Command-K to open the Sketch folder.

// All GML files must be located in the "data" folder for this sketch. 
// If you're not familiar with the Processing "data" folder, see here:
// http://processing.org/reference/environment/
// http://processing.org/learning/gettingstarted/ 

import processing.dxf.*;
Vector gmlFilenameVector;
int currentGMLFileIndex;
String gmlFilename;
Vector data;
PFont VagFont;

boolean bIsDxfRecording;
boolean bRotateTag90 = true;
boolean bIsTag3D = false;

// This number controls the maximum 'size' of your DXF output
// in 3D units -- for example, millimeters or inches, 
// as interpreted by your 3D program (Rhino etcetera).
// It also controls the display width of the preview.
float DXF_MAXIMUM_SIZE = 400.0; 

//=====================================================================
void setup(){

  size(800,600,P3D);

  println("Welcome to GML File Converter by Golan Levin.");
  bIsDxfRecording = false;
  findGMLFiles();
  //gmlFilename = "148.gml";

  loadData();
  normalizeData();

  VagFont = loadFont("VAGRounded-Lt-Normal-16.vlw"); 
  textFont(VagFont, 16);
  textMode(SCREEN);
}


//=====================================================================
// Identifies all GML files in the 'data' folder.
void findGMLFiles(){
  gmlFilenameVector = new Vector();
  String path = sketchPath + "/data";
  String[] filenames = listFileNames(path);
  for (int f=0; f<filenames.length; f++){
    if (filenames[f].endsWith(".gml")){
      println("Found GML file: " + filenames[f]);
      gmlFilenameVector.addElement(filenames[f]);
    }
  }

  currentGMLFileIndex = 0;
  if (gmlFilenameVector.size() > 0){
    gmlFilename = (String) gmlFilenameVector.elementAt(currentGMLFileIndex);
  }
}

//=====================================================================
// This function returns all the files in a directory as an array of Strings  
String[] listFileNames(String dir) {
  File file = new File(dir);
  if (file.isDirectory()) {
    String names[] = file.list();
    return names;
  } 
  else {
    // If it's not a directory
    return null;
  }
}


//=====================================================================
// Loads the data from the GML file into a Vector of GMLPoint arrays. 
// Each GMLPoint array is used to store the information of one stroke.
void loadData(){

  XMLElement gml = null;
  try{
    gml = new XMLElement(this, gmlFilename);
  } 
  catch (Exception e){
    println ("Problem loading GML file " + gmlFilename + "!");
  }

  boolean bPrintPointData = false;
  float tmp4[] = new float[4];
  String pointFieldNameArray[] = {
    "x", "y", "z", "time"        };

  if (gml != null){
    data = new Vector(10);
    int numCh0 = gml.getChildCount();
    if (numCh0 > 0){
      println("----------------------");
      println("Loading GML data from " + gmlFilename);

      for (int i=0; i<numCh0; i++) {
        XMLElement ch0 = gml.getChild(i);
        String ch0name = ch0.getName();
        if (ch0name.equals("tag")){
          XMLElement drawing = ch0.getChild("drawing");
          int numCh1 = drawing.getChildCount();
          int nStrokes = 0;

          for (int j=0; j<numCh1; j++){
            XMLElement ch1 = drawing.getChild(j);
            String ch1name = ch1.getName();
            if (ch1name.equals("stroke")){

              // Now we read in each stroke.
              int nPts = ch1.getChildCount();
              println("Loading data: stroke #" + j + " has " + nPts + " points.");
              GMLPoint pts[] = new GMLPoint[nPts];

              // For each point in the stroke
              for (int p=0; p<nPts; p++){

                // Fetch the p'th point of the stroke
                XMLElement pt = ch1.getChild(p);

                // Search for the field in that point which has the right name. 
                // Usually, these will be in the order "x", "y", "z", "time", 
                // But we need to do this search in case they're out of order for whatever reason.
                for (int n=0; n<4; n++){
                  XMLElement pointField = pt.getChild(n);
                  String pointFieldName = pointField.getName();
                  for (int m=0; m<4; m++){
                    if (pointFieldName.equals(pointFieldNameArray[m])){
                      tmp4[n] = Float.valueOf((pointField.getContent()).trim()).floatValue();
                      break;
                    }
                  }
                }
                float x = tmp4[0];
                float y = tmp4[1];
                float z = tmp4[2];
                float t = tmp4[3];
                pts[p] = new GMLPoint(x,y,z,t); 

                if (bPrintPointData){
                  println(p + " " + x + " " + y + " " + z + " " + t); 
                }
              }

              data.addElement(pts);

            }
          }
        }
      }
      println("Finished loading file."); 
    }
  }
}


//=====================================================================
// Normalizes all tag position data to the range 0...1
void normalizeData(){

  // Initialize range bounds
  float maxY = -99999;
  float maxX = -99999;
  float minY =  99999;
  float minX =  99999;

  // Compute the bounding ranges of the tag.
  // Not currently dealing with Z.
  for (int s=0; s<data.size(); s++){
    GMLPoint pts[] = (GMLPoint[]) data.elementAt(s);
    int nPts = pts.length;
    for (int p=0; p<nPts; p++){
      GMLPoint pt = pts[p];
      float x = pt.x;
      float y = pt.y;

      // While we're reading in the values, reckon the boundaries.
      if (y > maxY){
        maxY = y;
      }
      if (x > maxX){
        maxX = x;
      }
      if (y < minY){
        minY = y;
      }
      if (x < minX){
        minX = x;
      }
    }
  }

  // Normalize the marks to the range 0..1, 
  // without changing the aspect ratio of the tag.
  float rangeX = maxX - minX;
  float rangeY = maxY - minY;
  float normalizingRange = (rangeX > rangeY) ? rangeX : rangeY;
  float lesserRange      = (rangeX > rangeY) ? rangeY : rangeX;

  for (int s=0; s<data.size(); s++){
    GMLPoint pts[] = (GMLPoint[]) data.elementAt(s);
    int nPts = pts.length;
    for (int p=0; p<nPts; p++){
      GMLPoint pt = pts[p];
      pt.x = (pt.x - minX)/normalizingRange;
      pt.y = (pt.y - minY)/normalizingRange;
    }
  }
  minX /= normalizingRange;
  minY /= normalizingRange;
  maxX /= normalizingRange;
  maxY /= normalizingRange;

  if (bRotateTag90){
    minY =  99999;
    for (int s=0; s<data.size(); s++){
      GMLPoint pts[] = (GMLPoint[]) data.elementAt(s);
      int nPts = pts.length;
      for (int p=0; p<nPts; p++){
        GMLPoint pt = pts[p];
        float x = pt.x;
        float y = pt.y;
        pt.x = y;
        pt.y = 1.0 - x;

        if (pt.y < minY){
          minY = pt.y;
        }
      }
    }

    // realign
    for (int s=0; s<data.size(); s++){
      GMLPoint pts[] = (GMLPoint[]) data.elementAt(s);
      int nPts = pts.length;
      for (int p=0; p<nPts; p++){
        GMLPoint pt = pts[p];
        pt.y -= minY;
      }
    }
  }

  // At this point, the tag is guaranteed to be in the range 0...1
  // and touching both the XY axes. 
}


//=====================================================================
// Render to screen (and DXF if commanded)
void draw(){
  background(0);
  noFill();
  smooth();

  // Perform DXF export if so commanded
  if (bIsDxfRecording) {
    String dxfFileName = gmlFilename + ".dxf";//.txt";
    beginRaw (DXF, dxfFileName);
    stroke(0,0,0); // invisibility cheat
    renderTag();
    endRaw();
    bIsDxfRecording = false;
    println("\nSaved file " + dxfFileName);
  }
  
  // Display preview (Not file output):
  pushMatrix();
  float axisX = 100.0;
  float axisY = 100.0;
  stroke(64);
  line(axisX,0,axisX,height);
  line(0,axisY,width,axisY);
  translate(axisX,axisY,0);
  stroke(255,255,255);
  renderTag();
  popMatrix();

  fill(255);
  String instructions = "Current GML file: " + gmlFilename + "\n";
  instructions += "Click mouse for next GML file in data folder.\n";
  instructions += "Press 'd' to export DXF file to sketch folder.\n";
  instructions += "Press 'c' to export CSV file(s) to sketch folder.";
  text(instructions, width/2,height-80);
}

//=====================================================================
// draw the actual tag
void renderTag(){
  for (int s=0; s<data.size(); s++){
    GMLPoint pts[] = (GMLPoint[]) data.elementAt(s);
    beginShape();
    for (int p=0; p<(pts.length); p++){
      GMLPoint pt = pts[p];
      float x = pt.x * DXF_MAXIMUM_SIZE;
      float y = pt.y * DXF_MAXIMUM_SIZE;
      float z = (bIsTag3D) ? pt.z: 0.0;
      vertex(x,y,z);
    }
    endShape();
  }
}


//=====================================================================
// Export CSV files: one per stroke in the tag
void writeCsvFiles(){

  for (int s=0; s<data.size(); s++){
    GMLPoint pts[] = (GMLPoint[]) data.elementAt(s);
    int nPts = pts.length;
    String stringArray[] = new String[nPts];

    // populate array of strings
    for (int p=0; p<nPts; p++){
      GMLPoint pt = pts[p];
      float x = pt.x * DXF_MAXIMUM_SIZE;
      float y = pt.y * DXF_MAXIMUM_SIZE;
      String pointString = "" + nf(x,1,6) + ", " + nf(y,1,6)  ;
      if (p != (nPts-1)){
        pointString += ",";
      }
      stringArray[p] = pointString;
    }
    char aChar = (char)('a' + s);
    String csvFileName = gmlFilename + "_" + aChar + ".csv.txt";
    saveStrings(csvFileName, stringArray);
    println("Saved file " + csvFileName);
  }

}


//=====================================================================
void mousePressed(){
  // Whenever the user clicks the mouse, it will load the next GML file. 
  if (gmlFilenameVector.size() > 0){
    currentGMLFileIndex = (currentGMLFileIndex+1)%(gmlFilenameVector.size());
    gmlFilename = (String) gmlFilenameVector.elementAt(currentGMLFileIndex);
    loadData();
    normalizeData();
  }
}


//=====================================================================
// Based on keystrokes, issue commands to export DXF, etc.
void keyPressed(){
  switch(key){
  case 'C':
  case 'c':
    writeCsvFiles();
    break;
  case 'D':
  case 'd':
    bIsDxfRecording = true;
    break;
  case 'z':
  case 'Z':
    bIsTag3D = !bIsTag3D;
    break;
  } 
}


//=====================================================================
// A data structure for storing GML point data.
class GMLPoint {
  float x;
  float y;
  float z;
  float t;

  GMLPoint (float _x, float _y, float _z, float _t){
    x = _x;
    y = _y;
    z = _z;
    t = _t;
  }
}














